/**
 * Tarjan - Strongly Connected Components (SCC)
 *
 * Complexity (Time): O(V + E)
 * Complexity (Space): O(V + E)
 */

stack<int> S;
vector<int> graph[MAX];

int ncomp, ind;
int visited[MAX];

// The id of a node x (id[x]) is the order in which x was visited in the DFS
int id[MAX];

// The low-link value of a node x (low[x]) is defined as the smallest id
// of another node reachable from x when doing a DFS
int low[MAX];

// scc[i] contains the i-th SCC stored as a vector with the ids of the
// vertices in it
vector<int> scc[MAX];

/**
 * Fills SCC with strongly connected components of graph.
 * @param x any vertex
 */
void dfs(int x) {
  id[x] = low[x] = ind++;
  visited[x] = 1;

  S.push(x);

  for (auto i : graph[x])
    if (id[i] == -1) {
      dfs(i);

      // After the DFS, low[i] might have changed, in that case, low[x]
      // should be updated if a lower value has been found
      low[x] = min(low[x], low[i]);

    // If i has been visited, then i is "above" (has smaller id then x), 
    // meaning that the edge (x,i) is a "back edge"
    } else if (visited[i])
      low[x] = min(low[x], id[i]);

  // A SCC was found
  if (low[x] == id[x]) {
    int w;

    // Every node on top of x in the stack belongs to the SCC found
    do {
      w = S.top(); S.pop();
      visited[w] = 0;
      scc[ncomp].pb(w);
    } while (w != x);

    ncomp++;
  }
}

/**
 * Returns number of SCCs.
 * @param n number of vertices
 */
int tarjan(int n) {
  mset(id, -1);
  ncomp = ind = 0;

  for (int i = 0; i < n; ++i)
    scc[i].clear();

  // Apply tarjan in every component
  for (int i = 0; i < n; ++i)
    if (id[i] == -1)
      dfs(i);

  return ncomp;
}
